/*
 * Copyright (C) 2022 Ramiro Polla
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/*********************************************************************/
#include <pthread.h>
#include <SDL.h>

/*********************************************************************/
#include "libavutil/quickjs/quickjs-libc.h"
#include "libavutil/quickjs-sdl.h"
#include "internal.h"
#include "log.h"
#include "mem.h"

/*********************************************************************/
static const AVClass sdl_class = {
    .class_name = "quickjs-sdl",
    .item_name  = av_default_item_name,
    .version    = LIBAVUTIL_VERSION_INT,
};

typedef struct SDL_Event_Element {
    struct SDL_Event_Element *next;
    SDL_Event event;
} SDL_Event_Element;

typedef struct {
    const AVClass *class;
    pthread_mutex_t sdl_mutex;
    SDL_Joystick *joystick;
    SDL_GameController *gamecontroller;
    SDL_Event_Element *first_sdl_event;
    SDL_Event_Element *last_sdl_event;
} SDLContext;

/*********************************************************************/
static int get_SDL_Event(SDLContext *ctx, SDL_Event *event)
{
    SDL_Event_Element *element;
    int ret = 0;

    pthread_mutex_lock(&ctx->sdl_mutex);

    element = ctx->first_sdl_event;
    if ( element != NULL )
    {
        memcpy(event, &element->event, sizeof(SDL_Event));
        ctx->first_sdl_event = element->next;
        av_freep(&element);
        ret = 1;
    }

    pthread_mutex_unlock(&ctx->sdl_mutex);

    return ret;
}

static void add_SDL_Event(SDLContext *ctx, const SDL_Event *event)
{
    SDL_Event_Element *element;

    pthread_mutex_lock(&ctx->sdl_mutex);

    element = (SDL_Event_Element *) av_malloc(sizeof(SDL_Event_Element));
    element->next = NULL;
    memcpy(&element->event, event, sizeof(SDL_Event));
    if ( ctx->first_sdl_event == NULL )
        ctx->first_sdl_event = element;
    else
        ctx->last_sdl_event->next = element;
    ctx->last_sdl_event = element;

    pthread_mutex_unlock(&ctx->sdl_mutex);
}

/*********************************************************************/
static pthread_mutex_t sdlctxs_lock = PTHREAD_MUTEX_INITIALIZER;
static SDLContext **sdlctxs;
static int       nb_sdlctxs;

int ff_quickjs_sdl_do_events(void)
{
    return (nb_sdlctxs > 0);
}

void ff_quickjs_sdl_event_add(const SDL_Event *event)
{
    pthread_mutex_lock(&sdlctxs_lock);
    for ( int i = 0; i < nb_sdlctxs; i++ )
        add_SDL_Event(sdlctxs[i], event);
    pthread_mutex_unlock(&sdlctxs_lock);
}

/*********************************************************************/
static JSClassID js_SDL_class_id;

static JSValue SDL_ctor(JSContext *ctx, JSValueConst new_target, int argc, JSValueConst *argv)
{
    JSValue obj = JS_NULL;
    if ( SDL_WasInit(SDL_INIT_EVERYTHING) != 0 )
    {
        SDLContext *sdlctx;
        int sdlctx_idx;

        /* using new_target to get the prototype is necessary when the
           class is extended. */
        JSValue proto = JS_GetPropertyStr(ctx, new_target, "prototype");
        if ( JS_IsException(proto) )
            return JS_EXCEPTION;
        obj = JS_NewObjectProtoClass(ctx, proto, js_SDL_class_id);
        JS_FreeValue(ctx, proto);
        if ( JS_IsException(obj) )
        {
            JS_FreeValue(ctx, obj);
            return JS_EXCEPTION;
        }

        sdlctx = (SDLContext *) av_mallocz(sizeof(SDLContext));
        sdlctx->class = &sdl_class;
        pthread_mutex_init(&sdlctx->sdl_mutex, NULL);
        JS_SetOpaque(obj, sdlctx);

        /* add to sdlctxs */
        pthread_mutex_lock(&sdlctxs_lock);
        sdlctx_idx = nb_sdlctxs;
        sdlctxs = av_realloc_array(sdlctxs, ++nb_sdlctxs, sizeof(*sdlctxs));
        sdlctxs[sdlctx_idx] = sdlctx;
        pthread_mutex_unlock(&sdlctxs_lock);
    }
    return obj;
}

static void js_SDL_finalizer(JSRuntime *rt, JSValue val)
{
    SDLContext *sdlctx = (SDLContext *) JS_GetOpaque(val, js_SDL_class_id);

    /* remove from sdlctxs */
    pthread_mutex_lock(&sdlctxs_lock);
    for ( int i = 0; i < nb_sdlctxs; i++ )
    {
        if ( sdlctxs[i] == sdlctx )
        {
            int ip1 = i + 1;
            if ( ip1 != nb_sdlctxs )
                memmove(&sdlctxs[i], &sdlctxs[ip1], (nb_sdlctxs-ip1) * sizeof(*sdlctxs));
            break;
        }
    }
    sdlctxs = av_realloc_array(sdlctxs, --nb_sdlctxs, sizeof(*sdlctxs));
    pthread_mutex_unlock(&sdlctxs_lock);

    /* destroy and free sdlctx */
    if ( sdlctx->joystick != NULL )
        SDL_JoystickClose(sdlctx->joystick);
    if ( sdlctx->gamecontroller != NULL )
        SDL_GameControllerClose(sdlctx->gamecontroller);
    pthread_mutex_destroy(&sdlctx->sdl_mutex);
    av_free(sdlctx);
}

/*********************************************************************/
static JSValue
js_SDL_NumJoysticks(JSContext *ctx, JSValueConst this_val, int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, SDL_NumJoysticks());
}

static JSValue
js_SDL_JoystickOpen(JSContext *ctx, JSValueConst this_val, int argc, JSValueConst *argv)
{
    SDLContext *sdlctx = (SDLContext *) JS_GetOpaque(this_val, js_SDL_class_id);
    JSValue jnum = argv[0];
    int num;

    if ( !JS_IsInt32(jnum) )
    {
        JS_ThrowTypeError(ctx, "SDL_JoystickOpen() takes an integer as argument");
        return JS_EXCEPTION;
    }

    num = JS_VALUE_GET_INT(jnum);

    sdlctx->joystick = SDL_JoystickOpen(num);

    return JS_TRUE;
}

static JSValue
js_SDL_JoystickClose(JSContext *ctx, JSValueConst this_val, int argc, JSValueConst *argv)
{
    SDLContext *sdlctx = (SDLContext *) JS_GetOpaque(this_val, js_SDL_class_id);

    if ( sdlctx->joystick != NULL )
        SDL_JoystickClose(sdlctx->joystick);
    sdlctx->joystick = NULL;

    return JS_NULL;
}

static JSValue
js_SDL_GameControllerOpen(JSContext *ctx, JSValueConst this_val, int argc, JSValueConst *argv)
{
    SDLContext *sdlctx = (SDLContext *) JS_GetOpaque(this_val, js_SDL_class_id);
    JSValue jnum = argv[0];
    int num;

    if ( !JS_IsInt32(jnum) )
    {
        JS_ThrowTypeError(ctx, "SDL_GameControllerOpen() takes an integer as argument");
        return JS_EXCEPTION;
    }

    num = JS_VALUE_GET_INT(jnum);

    sdlctx->gamecontroller = SDL_GameControllerOpen(num);

    return JS_TRUE;
}

static JSValue
js_SDL_GameControllerClose(JSContext *ctx, JSValueConst this_val, int argc, JSValueConst *argv)
{
    SDLContext *sdlctx = (SDLContext *) JS_GetOpaque(this_val, js_SDL_class_id);

    if ( sdlctx->gamecontroller != NULL )
        SDL_GameControllerClose(sdlctx->gamecontroller);
    sdlctx->gamecontroller = NULL;

    return JS_NULL;
}

static JSValue
js_SDL_GetEvent(JSContext *ctx, JSValueConst this_val, int argc, JSValueConst *argv)
{
    SDLContext *sdlctx = (SDLContext *) JS_GetOpaque(this_val, js_SDL_class_id);
    SDL_Event event;
    JSValue jevent = JS_NULL;

    if ( get_SDL_Event(sdlctx, &event) )
    {
        /* prepare args */
        jevent = JS_NewObject(ctx);
        JS_SetPropertyStr(ctx, jevent, "type", JS_NewInt32(ctx, event.type));
        JS_SetPropertyStr(ctx, jevent, "timestamp", JS_NewInt32(ctx, event.common.timestamp));
        switch ( event.type )
        {
        case SDL_KEYDOWN:
        case SDL_KEYUP:
            {
                JSValue jkeysym = JS_NewObject(ctx);
                JS_SetPropertyStr(ctx, jevent, "state", JS_NewInt32(ctx, event.key.state));
                JS_SetPropertyStr(ctx, jevent, "repeat", JS_NewInt32(ctx, event.key.repeat));
                JS_SetPropertyStr(ctx, jevent, "keysym", jkeysym);
                JS_SetPropertyStr(ctx, jkeysym, "scancode", JS_NewInt32(ctx, event.key.keysym.scancode));
                JS_SetPropertyStr(ctx, jkeysym, "sym", JS_NewInt32(ctx, event.key.keysym.sym));
                JS_SetPropertyStr(ctx, jkeysym, "mod", JS_NewInt32(ctx, event.key.keysym.mod));
            }
            break;
        case SDL_JOYAXISMOTION:
            JS_SetPropertyStr(ctx, jevent, "which", JS_NewInt32(ctx, event.jaxis.which));
            JS_SetPropertyStr(ctx, jevent, "axis", JS_NewInt32(ctx, event.jaxis.axis));
            JS_SetPropertyStr(ctx, jevent, "value", JS_NewInt32(ctx, event.jaxis.value));
            break;
        case SDL_JOYBALLMOTION:
            JS_SetPropertyStr(ctx, jevent, "which", JS_NewInt32(ctx, event.jball.which));
            JS_SetPropertyStr(ctx, jevent, "ball", JS_NewInt32(ctx, event.jball.ball));
            JS_SetPropertyStr(ctx, jevent, "xrel", JS_NewInt32(ctx, event.jball.xrel));
            JS_SetPropertyStr(ctx, jevent, "yrel", JS_NewInt32(ctx, event.jball.yrel));
            break;
        case SDL_JOYHATMOTION:
            JS_SetPropertyStr(ctx, jevent, "which", JS_NewInt32(ctx, event.jhat.which));
            JS_SetPropertyStr(ctx, jevent, "hat", JS_NewInt32(ctx, event.jhat.hat));
            JS_SetPropertyStr(ctx, jevent, "value", JS_NewInt32(ctx, event.jhat.value));
            break;
        case SDL_JOYBUTTONDOWN:
        case SDL_JOYBUTTONUP:
            JS_SetPropertyStr(ctx, jevent, "which", JS_NewInt32(ctx, event.jbutton.which));
            JS_SetPropertyStr(ctx, jevent, "button", JS_NewInt32(ctx, event.jbutton.button));
            JS_SetPropertyStr(ctx, jevent, "state", JS_NewInt32(ctx, event.jbutton.state));
            break;
        case SDL_CONTROLLERAXISMOTION:
            JS_SetPropertyStr(ctx, jevent, "which", JS_NewInt32(ctx, event.caxis.which));
            JS_SetPropertyStr(ctx, jevent, "axis", JS_NewInt32(ctx, event.caxis.axis));
            JS_SetPropertyStr(ctx, jevent, "value", JS_NewInt32(ctx, event.caxis.value));
            break;
        case SDL_CONTROLLERBUTTONDOWN:
        case SDL_CONTROLLERBUTTONUP:
            JS_SetPropertyStr(ctx, jevent, "which", JS_NewInt32(ctx, event.cbutton.which));
            JS_SetPropertyStr(ctx, jevent, "button", JS_NewInt32(ctx, event.cbutton.button));
            JS_SetPropertyStr(ctx, jevent, "state", JS_NewInt32(ctx, event.cbutton.state));
            break;
        }
    }

    return jevent;
}

/*********************************************************************/
static JSClassDef js_SDL_class = {
    "SDL",
    .finalizer = js_SDL_finalizer,
};

static const JSCFunctionListEntry js_SDL_proto_funcs[] = {
    JS_CFUNC_DEF("numJoysticks", 0, js_SDL_NumJoysticks),
    JS_CFUNC_DEF("joystickOpen", 1, js_SDL_JoystickOpen),
    JS_CFUNC_DEF("joystickClose", 0, js_SDL_JoystickClose),
    JS_CFUNC_DEF("gameControllerOpen", 1, js_SDL_GameControllerOpen),
    JS_CFUNC_DEF("gameControllerClose", 0, js_SDL_GameControllerClose),
    JS_CFUNC_DEF("getEvent", 0, js_SDL_GetEvent),
};

static const JSCFunctionListEntry js_SDL_obj[] = {
    /* events types */
    JS_PROP_INT32_DEF("SDL_KEYDOWN", SDL_KEYDOWN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_KEYUP", SDL_KEYUP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_JOYAXISMOTION", SDL_JOYAXISMOTION, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_JOYBALLMOTION", SDL_JOYBALLMOTION, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_JOYHATMOTION", SDL_JOYHATMOTION, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_JOYBUTTONDOWN", SDL_JOYBUTTONDOWN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_JOYBUTTONUP", SDL_JOYBUTTONUP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_CONTROLLERAXISMOTION", SDL_CONTROLLERAXISMOTION, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_CONTROLLERBUTTONDOWN", SDL_CONTROLLERBUTTONDOWN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_CONTROLLERBUTTONUP", SDL_CONTROLLERBUTTONUP, JS_PROP_CONFIGURABLE),
    /* keycodes */
    JS_PROP_INT32_DEF("SDLK_UNKNOWN", SDLK_UNKNOWN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_RETURN", SDLK_RETURN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_ESCAPE", SDLK_ESCAPE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_BACKSPACE", SDLK_BACKSPACE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_TAB", SDLK_TAB, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_SPACE", SDLK_SPACE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_EXCLAIM", SDLK_EXCLAIM, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_QUOTEDBL", SDLK_QUOTEDBL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_HASH", SDLK_HASH, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_PERCENT", SDLK_PERCENT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_DOLLAR", SDLK_DOLLAR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_AMPERSAND", SDLK_AMPERSAND, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_QUOTE", SDLK_QUOTE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_LEFTPAREN", SDLK_LEFTPAREN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_RIGHTPAREN", SDLK_RIGHTPAREN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_ASTERISK", SDLK_ASTERISK, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_PLUS", SDLK_PLUS, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_COMMA", SDLK_COMMA, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_MINUS", SDLK_MINUS, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_PERIOD", SDLK_PERIOD, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_SLASH", SDLK_SLASH, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_0", SDLK_0, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_1", SDLK_1, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_2", SDLK_2, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_3", SDLK_3, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_4", SDLK_4, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_5", SDLK_5, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_6", SDLK_6, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_7", SDLK_7, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_8", SDLK_8, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_9", SDLK_9, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_COLON", SDLK_COLON, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_SEMICOLON", SDLK_SEMICOLON, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_LESS", SDLK_LESS, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_EQUALS", SDLK_EQUALS, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_GREATER", SDLK_GREATER, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_QUESTION", SDLK_QUESTION, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_AT", SDLK_AT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_LEFTBRACKET", SDLK_LEFTBRACKET, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_BACKSLASH", SDLK_BACKSLASH, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_RIGHTBRACKET", SDLK_RIGHTBRACKET, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_CARET", SDLK_CARET, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_UNDERSCORE", SDLK_UNDERSCORE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_BACKQUOTE", SDLK_BACKQUOTE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_a", SDLK_a, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_b", SDLK_b, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_c", SDLK_c, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_d", SDLK_d, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_e", SDLK_e, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_f", SDLK_f, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_g", SDLK_g, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_h", SDLK_h, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_i", SDLK_i, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_j", SDLK_j, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_k", SDLK_k, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_l", SDLK_l, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_m", SDLK_m, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_n", SDLK_n, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_o", SDLK_o, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_p", SDLK_p, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_q", SDLK_q, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_r", SDLK_r, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_s", SDLK_s, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_t", SDLK_t, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_u", SDLK_u, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_v", SDLK_v, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_w", SDLK_w, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_x", SDLK_x, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_y", SDLK_y, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_z", SDLK_z, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_CAPSLOCK", SDLK_CAPSLOCK, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F1", SDLK_F1, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F2", SDLK_F2, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F3", SDLK_F3, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F4", SDLK_F4, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F5", SDLK_F5, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F6", SDLK_F6, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F7", SDLK_F7, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F8", SDLK_F8, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F9", SDLK_F9, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F10", SDLK_F10, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F11", SDLK_F11, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F12", SDLK_F12, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_PRINTSCREEN", SDLK_PRINTSCREEN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_SCROLLLOCK", SDLK_SCROLLLOCK, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_PAUSE", SDLK_PAUSE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_INSERT", SDLK_INSERT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_HOME", SDLK_HOME, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_PAGEUP", SDLK_PAGEUP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_DELETE", SDLK_DELETE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_END", SDLK_END, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_PAGEDOWN", SDLK_PAGEDOWN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_RIGHT", SDLK_RIGHT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_LEFT", SDLK_LEFT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_DOWN", SDLK_DOWN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_UP", SDLK_UP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_NUMLOCKCLEAR", SDLK_NUMLOCKCLEAR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_DIVIDE", SDLK_KP_DIVIDE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_MULTIPLY", SDLK_KP_MULTIPLY, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_MINUS", SDLK_KP_MINUS, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_PLUS", SDLK_KP_PLUS, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_ENTER", SDLK_KP_ENTER, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_1", SDLK_KP_1, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_2", SDLK_KP_2, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_3", SDLK_KP_3, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_4", SDLK_KP_4, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_5", SDLK_KP_5, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_6", SDLK_KP_6, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_7", SDLK_KP_7, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_8", SDLK_KP_8, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_9", SDLK_KP_9, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_0", SDLK_KP_0, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_PERIOD", SDLK_KP_PERIOD, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_APPLICATION", SDLK_APPLICATION, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_POWER", SDLK_POWER, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_EQUALS", SDLK_KP_EQUALS, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F13", SDLK_F13, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F14", SDLK_F14, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F15", SDLK_F15, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F16", SDLK_F16, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F17", SDLK_F17, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F18", SDLK_F18, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F19", SDLK_F19, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F20", SDLK_F20, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F21", SDLK_F21, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F22", SDLK_F22, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F23", SDLK_F23, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_F24", SDLK_F24, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_EXECUTE", SDLK_EXECUTE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_HELP", SDLK_HELP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_MENU", SDLK_MENU, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_SELECT", SDLK_SELECT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_STOP", SDLK_STOP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_AGAIN", SDLK_AGAIN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_UNDO", SDLK_UNDO, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_CUT", SDLK_CUT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_COPY", SDLK_COPY, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_PASTE", SDLK_PASTE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_FIND", SDLK_FIND, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_MUTE", SDLK_MUTE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_VOLUMEUP", SDLK_VOLUMEUP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_VOLUMEDOWN", SDLK_VOLUMEDOWN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_COMMA", SDLK_KP_COMMA, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_EQUALSAS400", SDLK_KP_EQUALSAS400, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_ALTERASE", SDLK_ALTERASE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_SYSREQ", SDLK_SYSREQ, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_CANCEL", SDLK_CANCEL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_CLEAR", SDLK_CLEAR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_PRIOR", SDLK_PRIOR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_RETURN2", SDLK_RETURN2, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_SEPARATOR", SDLK_SEPARATOR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_OUT", SDLK_OUT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_OPER", SDLK_OPER, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_CLEARAGAIN", SDLK_CLEARAGAIN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_CRSEL", SDLK_CRSEL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_EXSEL", SDLK_EXSEL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_00", SDLK_KP_00, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_000", SDLK_KP_000, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_THOUSANDSSEPARATOR", SDLK_THOUSANDSSEPARATOR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_DECIMALSEPARATOR", SDLK_DECIMALSEPARATOR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_CURRENCYUNIT", SDLK_CURRENCYUNIT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_CURRENCYSUBUNIT", SDLK_CURRENCYSUBUNIT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_LEFTPAREN", SDLK_KP_LEFTPAREN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_RIGHTPAREN", SDLK_KP_RIGHTPAREN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_LEFTBRACE", SDLK_KP_LEFTBRACE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_RIGHTBRACE", SDLK_KP_RIGHTBRACE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_TAB", SDLK_KP_TAB, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_BACKSPACE", SDLK_KP_BACKSPACE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_A", SDLK_KP_A, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_B", SDLK_KP_B, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_C", SDLK_KP_C, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_D", SDLK_KP_D, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_E", SDLK_KP_E, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_F", SDLK_KP_F, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_XOR", SDLK_KP_XOR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_POWER", SDLK_KP_POWER, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_PERCENT", SDLK_KP_PERCENT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_LESS", SDLK_KP_LESS, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_GREATER", SDLK_KP_GREATER, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_AMPERSAND", SDLK_KP_AMPERSAND, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_DBLAMPERSAND", SDLK_KP_DBLAMPERSAND, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_VERTICALBAR", SDLK_KP_VERTICALBAR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_DBLVERTICALBAR", SDLK_KP_DBLVERTICALBAR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_COLON", SDLK_KP_COLON, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_HASH", SDLK_KP_HASH, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_SPACE", SDLK_KP_SPACE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_AT", SDLK_KP_AT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_EXCLAM", SDLK_KP_EXCLAM, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_MEMSTORE", SDLK_KP_MEMSTORE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_MEMRECALL", SDLK_KP_MEMRECALL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_MEMCLEAR", SDLK_KP_MEMCLEAR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_MEMADD", SDLK_KP_MEMADD, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_MEMSUBTRACT", SDLK_KP_MEMSUBTRACT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_MEMMULTIPLY", SDLK_KP_MEMMULTIPLY, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_MEMDIVIDE", SDLK_KP_MEMDIVIDE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_PLUSMINUS", SDLK_KP_PLUSMINUS, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_CLEAR", SDLK_KP_CLEAR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_CLEARENTRY", SDLK_KP_CLEARENTRY, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_BINARY", SDLK_KP_BINARY, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_OCTAL", SDLK_KP_OCTAL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_DECIMAL", SDLK_KP_DECIMAL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KP_HEXADECIMAL", SDLK_KP_HEXADECIMAL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_LCTRL", SDLK_LCTRL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_LSHIFT", SDLK_LSHIFT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_LALT", SDLK_LALT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_LGUI", SDLK_LGUI, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_RCTRL", SDLK_RCTRL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_RSHIFT", SDLK_RSHIFT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_RALT", SDLK_RALT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_RGUI", SDLK_RGUI, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_MODE", SDLK_MODE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_AUDIONEXT", SDLK_AUDIONEXT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_AUDIOPREV", SDLK_AUDIOPREV, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_AUDIOSTOP", SDLK_AUDIOSTOP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_AUDIOPLAY", SDLK_AUDIOPLAY, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_AUDIOMUTE", SDLK_AUDIOMUTE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_MEDIASELECT", SDLK_MEDIASELECT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_WWW", SDLK_WWW, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_MAIL", SDLK_MAIL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_CALCULATOR", SDLK_CALCULATOR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_COMPUTER", SDLK_COMPUTER, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_AC_SEARCH", SDLK_AC_SEARCH, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_AC_HOME", SDLK_AC_HOME, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_AC_BACK", SDLK_AC_BACK, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_AC_FORWARD", SDLK_AC_FORWARD, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_AC_STOP", SDLK_AC_STOP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_AC_REFRESH", SDLK_AC_REFRESH, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_AC_BOOKMARKS", SDLK_AC_BOOKMARKS, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_BRIGHTNESSDOWN", SDLK_BRIGHTNESSDOWN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_BRIGHTNESSUP", SDLK_BRIGHTNESSUP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_DISPLAYSWITCH", SDLK_DISPLAYSWITCH, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KBDILLUMTOGGLE", SDLK_KBDILLUMTOGGLE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KBDILLUMDOWN", SDLK_KBDILLUMDOWN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_KBDILLUMUP", SDLK_KBDILLUMUP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_EJECT", SDLK_EJECT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDLK_SLEEP", SDLK_SLEEP, JS_PROP_CONFIGURABLE),
    /* scancodes */
    JS_PROP_INT32_DEF("SDL_SCANCODE_UNKNOWN", SDL_SCANCODE_UNKNOWN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_A", SDL_SCANCODE_A, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_B", SDL_SCANCODE_B, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_C", SDL_SCANCODE_C, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_D", SDL_SCANCODE_D, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_E", SDL_SCANCODE_E, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F", SDL_SCANCODE_F, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_G", SDL_SCANCODE_G, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_H", SDL_SCANCODE_H, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_I", SDL_SCANCODE_I, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_J", SDL_SCANCODE_J, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_K", SDL_SCANCODE_K, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_L", SDL_SCANCODE_L, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_M", SDL_SCANCODE_M, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_N", SDL_SCANCODE_N, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_O", SDL_SCANCODE_O, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_P", SDL_SCANCODE_P, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_Q", SDL_SCANCODE_Q, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_R", SDL_SCANCODE_R, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_S", SDL_SCANCODE_S, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_T", SDL_SCANCODE_T, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_U", SDL_SCANCODE_U, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_V", SDL_SCANCODE_V, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_W", SDL_SCANCODE_W, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_X", SDL_SCANCODE_X, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_Y", SDL_SCANCODE_Y, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_Z", SDL_SCANCODE_Z, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_1", SDL_SCANCODE_1, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_2", SDL_SCANCODE_2, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_3", SDL_SCANCODE_3, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_4", SDL_SCANCODE_4, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_5", SDL_SCANCODE_5, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_6", SDL_SCANCODE_6, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_7", SDL_SCANCODE_7, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_8", SDL_SCANCODE_8, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_9", SDL_SCANCODE_9, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_0", SDL_SCANCODE_0, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_RETURN", SDL_SCANCODE_RETURN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_ESCAPE", SDL_SCANCODE_ESCAPE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_BACKSPACE", SDL_SCANCODE_BACKSPACE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_TAB", SDL_SCANCODE_TAB, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_SPACE", SDL_SCANCODE_SPACE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_MINUS", SDL_SCANCODE_MINUS, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_EQUALS", SDL_SCANCODE_EQUALS, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_LEFTBRACKET", SDL_SCANCODE_LEFTBRACKET, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_RIGHTBRACKET", SDL_SCANCODE_RIGHTBRACKET, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_BACKSLASH", SDL_SCANCODE_BACKSLASH, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_NONUSHASH", SDL_SCANCODE_NONUSHASH, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_SEMICOLON", SDL_SCANCODE_SEMICOLON, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_APOSTROPHE", SDL_SCANCODE_APOSTROPHE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_GRAVE", SDL_SCANCODE_GRAVE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_COMMA", SDL_SCANCODE_COMMA, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_PERIOD", SDL_SCANCODE_PERIOD, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_SLASH", SDL_SCANCODE_SLASH, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_CAPSLOCK", SDL_SCANCODE_CAPSLOCK, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F1", SDL_SCANCODE_F1, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F2", SDL_SCANCODE_F2, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F3", SDL_SCANCODE_F3, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F4", SDL_SCANCODE_F4, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F5", SDL_SCANCODE_F5, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F6", SDL_SCANCODE_F6, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F7", SDL_SCANCODE_F7, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F8", SDL_SCANCODE_F8, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F9", SDL_SCANCODE_F9, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F10", SDL_SCANCODE_F10, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F11", SDL_SCANCODE_F11, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F12", SDL_SCANCODE_F12, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_PRINTSCREEN", SDL_SCANCODE_PRINTSCREEN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_SCROLLLOCK", SDL_SCANCODE_SCROLLLOCK, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_PAUSE", SDL_SCANCODE_PAUSE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_INSERT", SDL_SCANCODE_INSERT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_HOME", SDL_SCANCODE_HOME, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_PAGEUP", SDL_SCANCODE_PAGEUP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_DELETE", SDL_SCANCODE_DELETE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_END", SDL_SCANCODE_END, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_PAGEDOWN", SDL_SCANCODE_PAGEDOWN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_RIGHT", SDL_SCANCODE_RIGHT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_LEFT", SDL_SCANCODE_LEFT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_DOWN", SDL_SCANCODE_DOWN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_UP", SDL_SCANCODE_UP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_NUMLOCKCLEAR", SDL_SCANCODE_NUMLOCKCLEAR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_DIVIDE", SDL_SCANCODE_KP_DIVIDE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_MULTIPLY", SDL_SCANCODE_KP_MULTIPLY, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_MINUS", SDL_SCANCODE_KP_MINUS, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_PLUS", SDL_SCANCODE_KP_PLUS, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_ENTER", SDL_SCANCODE_KP_ENTER, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_1", SDL_SCANCODE_KP_1, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_2", SDL_SCANCODE_KP_2, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_3", SDL_SCANCODE_KP_3, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_4", SDL_SCANCODE_KP_4, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_5", SDL_SCANCODE_KP_5, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_6", SDL_SCANCODE_KP_6, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_7", SDL_SCANCODE_KP_7, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_8", SDL_SCANCODE_KP_8, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_9", SDL_SCANCODE_KP_9, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_0", SDL_SCANCODE_KP_0, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_PERIOD", SDL_SCANCODE_KP_PERIOD, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_NONUSBACKSLASH", SDL_SCANCODE_NONUSBACKSLASH, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_APPLICATION", SDL_SCANCODE_APPLICATION, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_POWER", SDL_SCANCODE_POWER, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_EQUALS", SDL_SCANCODE_KP_EQUALS, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F13", SDL_SCANCODE_F13, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F14", SDL_SCANCODE_F14, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F15", SDL_SCANCODE_F15, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F16", SDL_SCANCODE_F16, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F17", SDL_SCANCODE_F17, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F18", SDL_SCANCODE_F18, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F19", SDL_SCANCODE_F19, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F20", SDL_SCANCODE_F20, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F21", SDL_SCANCODE_F21, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F22", SDL_SCANCODE_F22, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F23", SDL_SCANCODE_F23, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_F24", SDL_SCANCODE_F24, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_EXECUTE", SDL_SCANCODE_EXECUTE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_HELP", SDL_SCANCODE_HELP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_MENU", SDL_SCANCODE_MENU, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_SELECT", SDL_SCANCODE_SELECT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_STOP", SDL_SCANCODE_STOP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_AGAIN", SDL_SCANCODE_AGAIN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_UNDO", SDL_SCANCODE_UNDO, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_CUT", SDL_SCANCODE_CUT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_COPY", SDL_SCANCODE_COPY, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_PASTE", SDL_SCANCODE_PASTE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_FIND", SDL_SCANCODE_FIND, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_MUTE", SDL_SCANCODE_MUTE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_VOLUMEUP", SDL_SCANCODE_VOLUMEUP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_VOLUMEDOWN", SDL_SCANCODE_VOLUMEDOWN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_COMMA", SDL_SCANCODE_KP_COMMA, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_EQUALSAS400", SDL_SCANCODE_KP_EQUALSAS400, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_INTERNATIONAL1", SDL_SCANCODE_INTERNATIONAL1, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_INTERNATIONAL2", SDL_SCANCODE_INTERNATIONAL2, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_INTERNATIONAL3", SDL_SCANCODE_INTERNATIONAL3, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_INTERNATIONAL4", SDL_SCANCODE_INTERNATIONAL4, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_INTERNATIONAL5", SDL_SCANCODE_INTERNATIONAL5, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_INTERNATIONAL6", SDL_SCANCODE_INTERNATIONAL6, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_INTERNATIONAL7", SDL_SCANCODE_INTERNATIONAL7, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_INTERNATIONAL8", SDL_SCANCODE_INTERNATIONAL8, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_INTERNATIONAL9", SDL_SCANCODE_INTERNATIONAL9, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_LANG1", SDL_SCANCODE_LANG1, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_LANG2", SDL_SCANCODE_LANG2, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_LANG3", SDL_SCANCODE_LANG3, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_LANG4", SDL_SCANCODE_LANG4, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_LANG5", SDL_SCANCODE_LANG5, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_LANG6", SDL_SCANCODE_LANG6, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_LANG7", SDL_SCANCODE_LANG7, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_LANG8", SDL_SCANCODE_LANG8, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_LANG9", SDL_SCANCODE_LANG9, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_ALTERASE", SDL_SCANCODE_ALTERASE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_SYSREQ", SDL_SCANCODE_SYSREQ, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_CANCEL", SDL_SCANCODE_CANCEL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_CLEAR", SDL_SCANCODE_CLEAR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_PRIOR", SDL_SCANCODE_PRIOR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_RETURN2", SDL_SCANCODE_RETURN2, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_SEPARATOR", SDL_SCANCODE_SEPARATOR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_OUT", SDL_SCANCODE_OUT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_OPER", SDL_SCANCODE_OPER, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_CLEARAGAIN", SDL_SCANCODE_CLEARAGAIN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_CRSEL", SDL_SCANCODE_CRSEL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_EXSEL", SDL_SCANCODE_EXSEL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_00", SDL_SCANCODE_KP_00, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_000", SDL_SCANCODE_KP_000, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_THOUSANDSSEPARATOR", SDL_SCANCODE_THOUSANDSSEPARATOR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_DECIMALSEPARATOR", SDL_SCANCODE_DECIMALSEPARATOR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_CURRENCYUNIT", SDL_SCANCODE_CURRENCYUNIT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_CURRENCYSUBUNIT", SDL_SCANCODE_CURRENCYSUBUNIT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_LEFTPAREN", SDL_SCANCODE_KP_LEFTPAREN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_RIGHTPAREN", SDL_SCANCODE_KP_RIGHTPAREN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_LEFTBRACE", SDL_SCANCODE_KP_LEFTBRACE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_RIGHTBRACE", SDL_SCANCODE_KP_RIGHTBRACE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_TAB", SDL_SCANCODE_KP_TAB, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_BACKSPACE", SDL_SCANCODE_KP_BACKSPACE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_A", SDL_SCANCODE_KP_A, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_B", SDL_SCANCODE_KP_B, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_C", SDL_SCANCODE_KP_C, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_D", SDL_SCANCODE_KP_D, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_E", SDL_SCANCODE_KP_E, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_F", SDL_SCANCODE_KP_F, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_XOR", SDL_SCANCODE_KP_XOR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_POWER", SDL_SCANCODE_KP_POWER, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_PERCENT", SDL_SCANCODE_KP_PERCENT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_LESS", SDL_SCANCODE_KP_LESS, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_GREATER", SDL_SCANCODE_KP_GREATER, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_AMPERSAND", SDL_SCANCODE_KP_AMPERSAND, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_DBLAMPERSAND", SDL_SCANCODE_KP_DBLAMPERSAND, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_VERTICALBAR", SDL_SCANCODE_KP_VERTICALBAR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_DBLVERTICALBAR", SDL_SCANCODE_KP_DBLVERTICALBAR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_COLON", SDL_SCANCODE_KP_COLON, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_HASH", SDL_SCANCODE_KP_HASH, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_SPACE", SDL_SCANCODE_KP_SPACE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_AT", SDL_SCANCODE_KP_AT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_EXCLAM", SDL_SCANCODE_KP_EXCLAM, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_MEMSTORE", SDL_SCANCODE_KP_MEMSTORE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_MEMRECALL", SDL_SCANCODE_KP_MEMRECALL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_MEMCLEAR", SDL_SCANCODE_KP_MEMCLEAR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_MEMADD", SDL_SCANCODE_KP_MEMADD, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_MEMSUBTRACT", SDL_SCANCODE_KP_MEMSUBTRACT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_MEMMULTIPLY", SDL_SCANCODE_KP_MEMMULTIPLY, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_MEMDIVIDE", SDL_SCANCODE_KP_MEMDIVIDE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_PLUSMINUS", SDL_SCANCODE_KP_PLUSMINUS, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_CLEAR", SDL_SCANCODE_KP_CLEAR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_CLEARENTRY", SDL_SCANCODE_KP_CLEARENTRY, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_BINARY", SDL_SCANCODE_KP_BINARY, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_OCTAL", SDL_SCANCODE_KP_OCTAL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_DECIMAL", SDL_SCANCODE_KP_DECIMAL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KP_HEXADECIMAL", SDL_SCANCODE_KP_HEXADECIMAL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_LCTRL", SDL_SCANCODE_LCTRL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_LSHIFT", SDL_SCANCODE_LSHIFT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_LALT", SDL_SCANCODE_LALT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_LGUI", SDL_SCANCODE_LGUI, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_RCTRL", SDL_SCANCODE_RCTRL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_RSHIFT", SDL_SCANCODE_RSHIFT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_RALT", SDL_SCANCODE_RALT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_RGUI", SDL_SCANCODE_RGUI, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_MODE", SDL_SCANCODE_MODE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_AUDIONEXT", SDL_SCANCODE_AUDIONEXT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_AUDIOPREV", SDL_SCANCODE_AUDIOPREV, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_AUDIOSTOP", SDL_SCANCODE_AUDIOSTOP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_AUDIOPLAY", SDL_SCANCODE_AUDIOPLAY, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_AUDIOMUTE", SDL_SCANCODE_AUDIOMUTE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_MEDIASELECT", SDL_SCANCODE_MEDIASELECT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_WWW", SDL_SCANCODE_WWW, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_MAIL", SDL_SCANCODE_MAIL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_CALCULATOR", SDL_SCANCODE_CALCULATOR, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_COMPUTER", SDL_SCANCODE_COMPUTER, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_AC_SEARCH", SDL_SCANCODE_AC_SEARCH, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_AC_HOME", SDL_SCANCODE_AC_HOME, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_AC_BACK", SDL_SCANCODE_AC_BACK, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_AC_FORWARD", SDL_SCANCODE_AC_FORWARD, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_AC_STOP", SDL_SCANCODE_AC_STOP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_AC_REFRESH", SDL_SCANCODE_AC_REFRESH, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_AC_BOOKMARKS", SDL_SCANCODE_AC_BOOKMARKS, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_BRIGHTNESSDOWN", SDL_SCANCODE_BRIGHTNESSDOWN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_BRIGHTNESSUP", SDL_SCANCODE_BRIGHTNESSUP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_DISPLAYSWITCH", SDL_SCANCODE_DISPLAYSWITCH, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KBDILLUMTOGGLE", SDL_SCANCODE_KBDILLUMTOGGLE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KBDILLUMDOWN", SDL_SCANCODE_KBDILLUMDOWN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_KBDILLUMUP", SDL_SCANCODE_KBDILLUMUP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_EJECT", SDL_SCANCODE_EJECT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_SLEEP", SDL_SCANCODE_SLEEP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_APP1", SDL_SCANCODE_APP1, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_SCANCODE_APP2", SDL_SCANCODE_APP2, JS_PROP_CONFIGURABLE),
    /* keymods */
    JS_PROP_INT32_DEF("KMOD_NONE", KMOD_NONE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("KMOD_LSHIFT", KMOD_LSHIFT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("KMOD_RSHIFT", KMOD_RSHIFT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("KMOD_LCTRL", KMOD_LCTRL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("KMOD_RCTRL", KMOD_RCTRL, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("KMOD_LALT", KMOD_LALT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("KMOD_RALT", KMOD_RALT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("KMOD_LGUI", KMOD_LGUI, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("KMOD_RGUI", KMOD_RGUI, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("KMOD_NUM", KMOD_NUM, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("KMOD_CAPS", KMOD_CAPS, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("KMOD_MODE", KMOD_MODE, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("KMOD_RESERVED", KMOD_RESERVED, JS_PROP_CONFIGURABLE),
    /* joystick hat positions */
    JS_PROP_INT32_DEF("SDL_HAT_CENTERED", SDL_HAT_CENTERED, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_HAT_UP", SDL_HAT_UP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_HAT_RIGHT", SDL_HAT_RIGHT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_HAT_DOWN", SDL_HAT_DOWN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_HAT_LEFT", SDL_HAT_LEFT, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_HAT_RIGHTUP", SDL_HAT_RIGHTUP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_HAT_RIGHTDOWN", SDL_HAT_RIGHTDOWN, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_HAT_LEFTUP", SDL_HAT_LEFTUP, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_HAT_LEFTDOWN", SDL_HAT_LEFTDOWN, JS_PROP_CONFIGURABLE),
    /* General keyboard/mouse state definitions */
    JS_PROP_INT32_DEF("SDL_RELEASED", SDL_RELEASED, JS_PROP_CONFIGURABLE),
    JS_PROP_INT32_DEF("SDL_PRESSED", SDL_PRESSED, JS_PROP_CONFIGURABLE),
};

static void init_SDL(JSContext *ctx, JSValueConst global_object)
{
    JSRuntime *rt = JS_GetRuntime(ctx);
    JSValue SDL_proto = JS_NewObject(ctx);
    JSValue jSDL_ctor = JS_NewCFunction2(ctx, SDL_ctor, "SDL", 3, JS_CFUNC_constructor, 0);
    JS_NewClassID(&js_SDL_class_id);
    JS_NewClass(rt, js_SDL_class_id, &js_SDL_class);
    JS_SetPropertyFunctionList(ctx, SDL_proto, js_SDL_proto_funcs, FF_ARRAY_ELEMS(js_SDL_proto_funcs));
    JS_SetConstructor(ctx, jSDL_ctor, SDL_proto);
    JS_SetPropertyFunctionList(ctx, jSDL_ctor, js_SDL_obj, FF_ARRAY_ELEMS(js_SDL_obj));
    JS_SetClassProto(ctx, js_SDL_class_id, SDL_proto);
    JS_SetPropertyStr(ctx, global_object, "SDL", jSDL_ctor);
}

/*********************************************************************/
void ff_quickjs_sdl_init(JSContext *ctx, JSValueConst global_object)
{
    init_SDL(ctx, global_object);
}
